/*
    GDAM - Geoff & Dave's Audio Mixer
    Copyright (C) 1999    Dave Benson, Geoff Matters.

    This library is free software; you can redistribute it and/or
    modify it under the terms of the GNU Lesser General Public
    License as published by the Free Software Foundation; either
    version 2 of the License, or (at your option) any later version.

    This library is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
    Lesser General Public License for more details.

    You should have received a copy of the GNU Lesser General Public
    License along with this library; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307  USA

    Contact:
        daveb@ffem.org <Dave Benson>
        geoff@ugcs.caltech.edu <Geoff Matters>
*/
#include <unistd.h>
#include <sys/stat.h>
#include <sys/types.h>
#include <linux/soundcard.h>
#include <stdio.h>
#include <fcntl.h>
#include <stdlib.h>
#include <sys/ioctl.h>
#include <errno.h>
#include "gdamdevicealsa.h"

#define GDAM_SOURCE_PRIORITY G_PRIORITY_DEFAULT

typedef struct _GdamDeviceOptions GdamDeviceOptions;

static gboolean    gdam_device_alsa_open      (GdamSink*         sink,
                                               const char*       name,
				               const char*       opts);
static void        gdam_device_alsa_close     (GdamSink*         sink);
static gboolean    gdam_device_alsa_set_source(GdamSink*         sink,
                                               GdamSource*       source);
static void        gdam_device_alsa_remove_source(GdamSink*      sink);
static int         gdam_device_alsa_write     (GdamSink*         output,
	                                       const GdamSample* samples,
					       int               max_read);
static void        gdam_device_alsa_destroy   (GtkObject*        to_kill);


static int         gdam_device_alsa_open_fd  (const char*        device_name,
                                              GdamDeviceOptions* device_opts);

/* Add/remove the sink from the main loop. */
static void        gdam_device_alsa_stop      (GdamDeviceAlsa*    device);
static void        gdam_device_alsa_play      (GdamDeviceAlsa*    device);

/* Functions passed to g_source_add. */
static gboolean    gdam_device_alsa_prepare   (gpointer          src_data,
                                               GTimeVal*         curtime,
                                               gint*             timeout,
		                               gpointer          user_data);
static gboolean    gdam_device_alsa_check     (gpointer          src_data,
                                               GTimeVal*         curtime,
		                               gpointer          user_data);
static gboolean    gdam_device_alsa_dispatch  (gpointer          src_data,
                                               GTimeVal*         curtime,
		                               gpointer          user_data);

/* And this function does a single fragment-sized write out to
 * the file descriptor. */
static gboolean    gdam_device_alsa_write_one_fragment
                                              (GdamDeviceAlsa*    device_alsa);


static void gdam_device_alsa_class_init(GdamSinkClass*  klass)
{
	GtkObjectClass* object_class = (GtkObjectClass*)klass;
	klass->write = gdam_device_alsa_write;
	klass->set_source = gdam_device_alsa_set_source;
	klass->remove_source = gdam_device_alsa_remove_source;
	klass->open = gdam_device_alsa_open;
	klass->close = gdam_device_alsa_close;
	object_class->destroy = gdam_device_alsa_destroy;
}

static void gdam_device_alsa_init(GdamDeviceAlsa*  sink)
{
	sink->device_fd = -1;
	sink->poll_fd.fd = -1;
	sink->source_id = 0;
	sink->fragment_size = (1 << 11);
}

static GtkObjectClass* the_sink_class;
static GtkType device_alsa_type = 0;
GtkType gdam_device_alsa_get_type()
{
	if (!device_alsa_type) {
		static const GtkTypeInfo sink_info = {
			"GdamDeviceAlsa",
			sizeof(GdamDeviceAlsa),
			sizeof(GdamSinkClass),
			(GtkClassInitFunc) gdam_device_alsa_class_init,
			(GtkObjectInitFunc) gdam_device_alsa_init,
			/* reserved_1*/ NULL,
			/* reserved_2*/ NULL,
			(GtkClassInitFunc) NULL
		};
		the_sink_class = gtk_type_class(GDAM_TYPE_SINK);
		device_alsa_type = gtk_type_unique(GDAM_TYPE_SINK, &sink_info);
	}
	return device_alsa_type;
}

GdamSink* gdam_device_alsa_new(gpointer      device_data,
                              const char*   device_name,
			      const char*   device_options)
{
	GtkType type = gdam_device_alsa_get_type();
	GdamDeviceAlsa* dev_alsa = (GdamDeviceAlsa*)gtk_type_new(type);
	if (!gdam_sink_open(GDAM_SINK(dev_alsa), device_name, device_options)) {
		gtk_object_unref(GTK_OBJECT(dev_alsa));
		return NULL;
	}
	return GDAM_SINK(dev_alsa);
}

#define GDAM_SAMPLE_FORMAT_NATIVE16	0
#define GDAM_SAMPLE_FORMAT_NATIVE8	1
struct _GdamDeviceOptions {
	int		rate;		/* samples per second. */
	gboolean	do_ioctls;	/* Whether or not to actually 
	                                 *    run the sound card ioctls.
					 * Needed to reuse this class for
					 *    raw output. */
	int		num_channels;	/* must be 2 for now. */
	int		format;

	int		num_fragments;
	int		log_frag_size;
	int             sample_size;		/* in bits. */
};


/*
 *                                  _            _
 *  _ __   __ _ _ __ ___  ___    __| | _____   _(_) ___ ___
 * | '_ \ / _` | '__/ __|/ _ \  / _` |/ _ \ \ / / |/ __/ _ \
 * | |_) | (_| | |  \__ \  __/ | (_| |  __/\ V /| | (_|  __/
 * | .__/ \__,_|_|  |___/\___|  \__,_|\___| \_/ |_|\___\___|
 * |_|
 *              _   _
 *   ___  _ __ | |_(_) ___  _ __  ___
 *  / _ \| '_ \| __| |/ _ \| '_ \/ __|
 * | (_) | |_) | |_| | (_) | | | \__ \
 *  \___/| .__/ \__|_|\___/|_| |_|___/
 *       |_|
 */
static gboolean gdam_device_parse_options(GdamDeviceOptions* options,
                                          const char*        string)
{
	char** opts;
	char** opts_at;
	options->rate = 44100;
	options->do_ioctls = TRUE;
	options->num_channels = 2;
	options->format = GDAM_SAMPLE_FORMAT_NATIVE16;
	options->num_fragments = 4;
	options->log_frag_size = 11;
	opts = g_strsplit(string, ":", 100);
	for (opts_at = opts; *opts_at; opts_at++) {
		char* opt = *opts_at;
		if (strncmp(opt, "logfragsize=", 12) == 0) {
			int log_frag_size = atoi(opt + 12);
			if (log_frag_size < 6 || log_frag_size > 100) {
				g_warning("log_frag_size was out of bounds...");
			} else {
				options->log_frag_size = log_frag_size;
			}
		} else if (strncmp(opt, "bits=", 5) == 0) {
			int bits = atoi(opt + 5);
			if (bits == 8)
				options->format = GDAM_SAMPLE_FORMAT_NATIVE8;
			else if (bits == 16)
				options->format = GDAM_SAMPLE_FORMAT_NATIVE16;
			else {
				g_warning("unknown bits/sample %d", bits);
			}
		} else if (strcmp(opt, "mono") == 0) {
			options->num_channels = 1;
		} else if (strcmp(opt, "stereo") == 0) {
			options->num_channels = 2;
		} else if (strncmp(opt, "rate=", 5) == 0) {
			options->rate = atoi(opt + 5);
		} else if (strncmp(opt, "numfragments=", 13) == 0) {
			int num_fragments = atoi(opt + 13);
			if (num_fragments < 3 || num_fragments > 300) {
				g_warning("numfragments was out of bounds...");
			} else {
				options->num_fragments = num_fragments;
			}
		} else {
			g_warning("only options known to alsa are "
				"numfragments and logfragsize [was %s]", opt);
		}
		g_free(opt);
	}
	g_free(opts);

	return TRUE;
}
static int bits_per_sample(int format)
{
	if (format == GDAM_SAMPLE_FORMAT_NATIVE16)	return 16;
	if (format == GDAM_SAMPLE_FORMAT_NATIVE8)	return 8;
	return -1;
}

static gboolean writen(int fd, const char* buf, int len)
{
        while (len > 0) {
                int n = write(fd, buf, len);
                if (n < 0) {
                        if (errno == EINTR)
                                continue;
                        return FALSE;
                }
                buf += n; len -= n;
        }
        return TRUE;
}

static gboolean readn(int fd, char* buf, int len)
{
        while (len > 0) {
                int n = read(fd, buf, len);
                if (n < 0) {
                        if (errno == EINTR)
                                continue;
                        return FALSE;
                }
                if (n == 0) return FALSE;
                buf += n; len -= n;
        }
        return TRUE;
}

static void*
gdam_device_alsa_open(const char*          device_name,
                      GdamDeviceOptions*   device_options)
{
	int fd;
	int t; unsigned long r;
	int log_frag_size = device_options->log_frag_size;
	int num_fragments = device_options->num_fragments;
	int stereo = 1, rate, sample_size, num_channels;
	snd_pct_t* snd_pcm;
	if (!device_name)
		device_name = "/dev/dsp";

	if (snd_pcm_open(&snd_pcm_handle, 
	                 CARD_NUMBER,
			 DEVICE_NUMBER,
			 SND_PCM_OPEN_PLAYBACK) < 0) {
		fprintf(stderr, "couldn't open sound device %s: %s\n",
			device_name, strerror(errno));
		return -1;
	}

	/* Close the device on exec. */
	fcntl(fd, F_SETFD, FD_CLOEXEC);

	TODO:
		- set the number of fragments
		- set sample size + rate
	gdam_debug("DSP Block size is %lu", r);
	/*sound_device_block_size = r;*/

	return snd_pcm_handle;
}

/* Open the sound device. */
static gboolean
gdam_device_alsa_open(GdamSink* sink, const char* name, const char* opts)
{
	GdamDeviceAlsa* dev_alsa = GDAM_DEVICE_ALSA(sink);
	int fd;
	GdamDeviceOptions options;

	if (!gdam_device_parse_options(&options, opts)) {
		g_warning("error parsing options `%s'", opts);
		return FALSE;
	}

	dev_alsa->fragment_size = (1 << options.log_frag_size);

	fd = gdam_device_alsa_open_fd(name, &options);
	if (fd < 0) {
		g_warning("error opening sound device %s", name);
		return FALSE;
	}
	dev_alsa->device_fd = fd;
	sink->device_name = g_strdup(name);
	gdam_debug("dev_alsa: sample_size=%d", options.sample_size);

	return TRUE;
}

/* Close the sound device. */
static void
gdam_device_alsa_close(GdamSink* sink)
{
	GdamDeviceAlsa* dev_alsa = GDAM_DEVICE_ALSA(sink);
	gdam_device_alsa_stop(dev_alsa);
	if (dev_alsa->device_fd >= 0) {
		close(dev_alsa->device_fd);
		dev_alsa->device_fd = -1;
	}
}

static gboolean
gdam_device_alsa_set_source(GdamSink* sink, GdamSource* source)
{
	sink->feed = source;
	gtk_object_ref(GTK_OBJECT(source));
	gdam_device_alsa_play(GDAM_DEVICE_ALSA(sink));
	return TRUE;
}

static void
gdam_device_alsa_remove_source(GdamSink* sink)
{
	gdam_device_alsa_stop(GDAM_DEVICE_ALSA(sink));
	gtk_object_unref(GTK_OBJECT(sink->feed));
}

static void gdam_device_alsa_destroy(GtkObject* to_kill)
{
	gdam_device_alsa_stop(GDAM_DEVICE_ALSA(to_kill));
	gdam_device_alsa_close(GDAM_SINK(to_kill));
}

static int         gdam_device_alsa_write     (GdamSink*         output,
	                                      const GdamSample* samples,
					      int               max_read)
{
	g_assert_not_reached();
	return -1;
}


/*
 * Adding and detaching from the main loop.
 */
static GSourceFuncs device_alsa_source_funcs = {
	gdam_device_alsa_prepare,
	gdam_device_alsa_check,
	gdam_device_alsa_dispatch,
	NULL
};
static void gdam_device_alsa_source_destroy_notify(gpointer user_data);

static void gdam_device_alsa_stop(GdamDeviceAlsa* device)
{
	gdam_debug("gdam_device_alsa_stop...");
	if (device->poll_fd.fd >= 0) {
		g_main_remove_poll(&device->poll_fd);
		device->poll_fd.fd = -1;
	}
	if (device->source_id != 0) {
		guint source_id = device->source_id;
		device->source_id = 0;
		g_source_remove(source_id);
	}
}

static void gdam_device_alsa_play(GdamDeviceAlsa* device)
{
	g_return_if_fail(device->device_fd >= 0);
	g_return_if_fail(device->source_id == 0);
	g_return_if_fail(device->poll_fd.fd == -1);

	device->poll_fd.fd = device->device_fd;
	device->poll_fd.events = G_IO_OUT;
	g_main_add_poll(&device->poll_fd, GDAM_SOURCE_PRIORITY);

	device->source_id = g_source_add(
		GDAM_SOURCE_PRIORITY,
		TRUE,
		&device_alsa_source_funcs,
		NULL,
		device,
		gdam_device_alsa_source_destroy_notify
	);
	g_assert(device->source_id != 0);
}

static void gdam_device_alsa_source_destroy_notify(gpointer user_data)
{
	GdamDeviceAlsa* dev_alsa = GDAM_DEVICE_ALSA(user_data);
	dev_alsa->source_id = 0;
	gdam_device_alsa_stop(dev_alsa);
}

static gboolean    gdam_device_alsa_prepare   (gpointer          src_data,
                                              GTimeVal*         curtime,
                                              gint*             timeout,
		                              gpointer          user_data)
{
	*timeout = -1;
	return FALSE;
}

static gboolean    gdam_device_alsa_check     (gpointer          src_data,
                                              GTimeVal*         curtime,
		                              gpointer          user_data)
{
	GdamDeviceAlsa* dev_alsa = (GdamDeviceAlsa*)user_data;
	return dev_alsa->poll_fd.revents == 0 ? FALSE : TRUE;
}

static gboolean    gdam_device_alsa_dispatch  (gpointer          src_data,
                                              GTimeVal*         curtime,
		                              gpointer          user_data)
{
	if (!gdam_device_alsa_write_one_fragment(GDAM_DEVICE_ALSA(user_data)))
		return FALSE;
	return TRUE;
}

static gboolean
gdam_device_alsa_write_one_fragment(GdamDeviceAlsa* device_alsa)
{
	char* fragment_buffer;
	int num_samples = device_alsa->fragment_size / sizeof(GdamSample);
	int num_read;
	GdamSink* sink = (GdamSink*)device_alsa;
	GdamSource* feed = sink->feed;

	/* The fragment size is the ideal amount to write to
	 * an alsa device atomically. (I think) */
	int fragment_size = device_alsa->fragment_size;
	int out_fragment_size = fragment_size / sizeof(GdamSample)
	                      * device_alsa->sample_size / 8;

	g_return_val_if_fail(feed != NULL, FALSE);

	fragment_buffer = (char*)alloca(fragment_size);


	/* Read, then miss. */
	if (!gdam_source_is_ready_to_read(feed))
		num_read = 0;
	else {
		num_read = gdam_source_read(feed,
		                            (GdamSample*)fragment_buffer,
					    num_samples);
		if (num_read <= 0) {
			gdam_debug("writing to device: subsource ended, "
			           "returned %d", num_read);

			/* Remove source sets feed to NULL. */
			gdam_sink_remove_source(sink);

			return FALSE;
		}
	}

	if (num_read < num_samples)
		gdam_source_miss_data(feed, num_samples - num_read);
	if (device_alsa->downsample_func) {
		int num_samples = fragment_size / sizeof(GdamSample);
		(*device_alsa->downsample_func)(fragment_buffer, num_samples);
	}

	for (;;) {
		int num_written = write(device_alsa->device_fd,
		                        fragment_buffer,
				        out_fragment_size);
		if (num_written < 0) {
			if (errno == EINTR)
				continue;
			if (errno == EWOULDBLOCK) {
				g_warning("TODO: " /*TODO*/
				          "not able to handle non-ready, "
					  "non-blocking situation "
					  "(SHOULD QUEUE DATA)");
				g_warning("ditching a fragment of data");
				return TRUE;
			}
			g_warning("error writing to sound device: %s",
				  strerror(errno));

			/* Will set device_fd to -1. */
			gdam_sink_close(sink);

			return FALSE;
		}
		
		if (num_written < out_fragment_size) {
			static gboolean one_shot = FALSE;
			if (!one_shot) {
				g_warning("hmmm... your sound device didn't "
					  "accept a whole fragment...");
				g_warning("something may be wrong, but "
					  "we'll proceed");
				one_shot = TRUE;
			}
			g_assert(num_written != 0);
			fragment_size -= num_written;
			fragment_buffer += num_written;
			continue;
		}
		break;
	}

	return TRUE;
}
